package benchmark;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * <code>Data</code> is a class for loading results files and representing in a organized data structure
 * 
 * @authors 
 * <a href="mailto:ezzahir@ensa-agadir.ac.ma?subject=DisChoco &cc=wahbi@lirmm.fr &cc=bessiere@lirmm.fr">
 * Redoaune Ezzahir</a> and 
 * <a href="mailto:wahbi@lirmm.fr?subject=DisChoco &cc=ezzahir@ensa-agadir.ac.ma &cc=bessiere@lirmm.fr">
 * Mohamed Wahbi</a>
 * 
 */
import util.Tools;

public class Data {
	public ArrayList<String[]> data = null;
	public HashMap<String, Integer> keysMap = null;

	Data() {
		this.data = new ArrayList<String[]>();
		this.setKeysMap(new HashMap<String, Integer>());
	}

	Data(ArrayList<String[]> data) {
		this.data = data;
	}

	public void addElement(String[] e) {
		this.data.add(this.data.size(), e);
	}

	public String[] getElement(int index) {
		if(index<this.data.size()){
			return this.data.get(index);
		}
		String[] ans = new String[this.data.get(0).length];
		for(int i=0 ; i<ans.length ; i++){
			ans[i] = "0";
		}
		return ans;
	}

	public String toString() {
		String ans = "";
		ans += this.keysMap + "\n";
		for (int l = 0 ; l < this.data.size() ; l++) {
			String[] line = this.data.get(l);
			for (int i = 0 ; i < line.length ; i++) {
				ans += line[i] + " ";
			}
			ans += "\n";
		}
		return ans;
	}

	public int getNbLine() {
		return this.data.size();
	}

	public void addKey(String key, Integer value) {
		this.keysMap.put(key, value);
	}

	public int getKey(String key) {
		if (this.keysMap.containsKey(key)) { return this.keysMap.get(key); }
		return -1;
	}

	public void setKeysMap(HashMap<String, Integer> keysMap) {
		this.keysMap = keysMap;
	}

	public void setKeys(String[] keys) {
		for (int k = 0 ; k < keys.length ; k++) {
			this.addKey(keys[k], k);
		}
	}

	public HashMap<String, Integer> getKeysMap() {
		return keysMap;
	}

	public String[] getKeys() {
		return Tools.dataKeyToString(this.keysMap).split(",");
	}

	public static Data fromFile(String fileName, int nbOfIgnoredLines) {
		Data data = new Data();
		try {
			String line = "";
			FileReader file = new FileReader(fileName);
			BufferedReader br = new BufferedReader(file);
			for (int l = 0 ; l < nbOfIgnoredLines ; l++) {
				line = br.readLine();
			}
			line = br.readLine();
			line = line.replaceAll(", ", ",");
			data.setKeys(line.split(","));
			while ((line = br.readLine()) != null) {
				line = line.replaceAll(", ", ",");
				data.addElement(line.split(","));
			}
			file.close();
		} catch (IOException e) {}
		return data;
	}

	public double[] getColumn(String key) {
		double[] column = new double[this.getNbLine()];
		for (int i = 0 ; i < column.length ; i++) {
			int c = this.getKey(key);
			if (c == -1) {
				column[i] = 0;
				continue;
			}
			String value = this.getElement(i)[c];
			column[i] = new Double(value);
		}
		return column;
	}

}
