/**
 * This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with this program; if not, write to
 * the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * @version 2.0 DisChoco GraphPanel.java
 */
package gui.visualizing;

import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import dischoco.DisChoco;

public class GraphPanel extends JPanel {

	private static final long serialVersionUID = -866986744241325876L;
	private Graph<Vertex, Edge> graph = null;
	private RenderingHints hints = null;
	private FontMetrics m = null;
	private Vertex pick = null;
	private boolean pickfixed;
	public static final int MAX_ZOOM = 5;
	private double forceConstant;
	private double temperature;
	private int currentIteration;
	private int maxIterations = 1000;
	private double attraction_multiplier = 0.75;
	private double attraction_constant;
	private double repulsion_multiplier = 0.75;
	private double repulsion_constant;
	private double max_dimension;
	private static final double EPSILON = 0.1D;
	private static final int BORDER = 32;
	private boolean PICKING = true;// transform mode false
	private static final String instructions =
		"<html>" + 
		"<b><h4><center> Instructions for Annotations </center></h4></b>" + 
		"<pre>" + 
		"To do." +
		"</pre>";
	public static final int NORTH_PANEL_HEIGHT = 78;
	private JPanel northPanel = null;
	private JPanel canvas = null;
	private JButton zoomIn = null;
	private JButton zoomOut = null;
	private Dimension preferredSize = null;
	private Dimension minimumSize = null;
	private Dimension maximumSize = null;
	private JComboBox modeBox = null;
	private JComboBox layoutBox = null;
	private JButton helpButton = null;
	private JDialog helpDialog = null;
	private JPanel modeControls = null;
	private JPanel layoutControls = null;
	private JPanel helpControls = null;
	private JPanel zoomControls;
	private Dimension dimension;

	public GraphPanel(Graph<Vertex, Edge> g, Dimension d) {
		super();
		hints = new RenderingHints(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		hints.add(new RenderingHints(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY));
		setSize(d);
		graph = g;
		setLayout(new BorderLayout());
		add(getNorthPanel(d), BorderLayout.NORTH);
		add(new JScrollPane(getCanvas(d)), BorderLayout.CENTER);
		setBorder(BorderFactory.createEtchedBorder());
		((Canvas) getCanvas(d)).scale(1.0f);
	}

	public void addEdge(Vertex v1, Vertex v2) {
		graph.addEdge(new Edge(v1, v2), new Pair<Vertex>(v1, v2));
	}

	/**
	 * This method initializes northPanel
	 * 
	 * @param height
	 * @param width
	 * @return javax.swing.JPanel
	 */
	public JPanel getNorthPanel(Dimension d) {
		if (northPanel == null) {
			northPanel = new JPanel();
			northPanel.setLayout(new FlowLayout());  
			northPanel.setPreferredSize(new Dimension(d.width, NORTH_PANEL_HEIGHT));
			northPanel.setMinimumSize(new Dimension(d.width, NORTH_PANEL_HEIGHT));
			northPanel.add(getZoomIn());
			northPanel.add(getZoomOut());
			northPanel.add(getZoomControls());
			northPanel.add(getModeControls());
			northPanel.add(getLayoutControls());
			northPanel.add(getHelpControls());
		}
		return northPanel;
	}

	/**
	 * @return JPanel
	 */
	public JPanel getZoomControls() {
		if(zoomControls == null){
			zoomControls = new JPanel();
			zoomControls.setBorder(BorderFactory.createTitledBorder("Zoom"));
			zoomControls.add(getZoomIn());
			zoomControls.add(getZoomOut());
		}
		return zoomControls;
	}

	/**
	 * @return JPanel
	 */
	public JPanel getModeControls() {
		if(modeControls == null){
			modeControls = new JPanel();
			modeControls.setBorder(BorderFactory.createTitledBorder("Mouse"));
			modeControls.add(getModeComboBox());	
		}
		return modeControls;
	}
	
	/**
	 * @return JPanel
	 */
	public JPanel getLayoutControls() {
		if(layoutControls == null){
			layoutControls = new JPanel();
			layoutControls.setBorder(BorderFactory.createTitledBorder("Layout"));
			layoutControls.add(getLayoutComboBox());
		}
		return layoutControls;
	}
	
	/** 
	 * @return JPanel
	 */
	public JPanel getHelpControls() {
		if(helpControls == null){
			helpControls = new JPanel();
			helpControls.setBorder(BorderFactory.createTitledBorder("Help"));
			helpControls.add(getHelpButton());
		}
		return helpControls;
	}
	
	public JButton getHelpButton() {
		if (helpButton == null) {
			helpButton = new JButton("Help");
			helpButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					getHelpDialog().pack();
					Container c = getParent();
					while(!(c instanceof JFrame)){
						c = c.getParent();
					}
					JFrame frame = (JFrame) c;
					Point loc = frame.getLocation();
					Dimension size = frame.getSize();
					Dimension d = helpDialog.getSize();
					helpDialog.setLocation(loc.x+(size.width-d.width)/2, loc.y + (size.height-d.height)/2);
					helpDialog.setVisible(true);
				}
			});
		}
		return helpButton;
	}

	private JDialog getHelpDialog() {
		if (helpDialog == null) {
			helpDialog = new JDialog();
			helpDialog.getContentPane().add(new JLabel(instructions));
		}
		return helpDialog;
	}

	/**
	 * @return Returns the modeBox.
	 */
	public JComboBox getModeComboBox() {
		if (modeBox == null) {
			modeBox = new JComboBox(new String[] {"PICKING", "TRANSFORMING",});
			modeBox.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent paramItemEvent) {
					if (modeBox.getSelectedItem().equals("PICKING")) {
						PICKING = true;
					} else {
						PICKING = false;
					}
				}
			});
		}
		modeBox.setSelectedItem("PICKING");
		return modeBox;
	}

	/**
	 * @return Returns the modeBox.
	 */
	public JComboBox getLayoutComboBox() {
		if (layoutBox == null) {
			layoutBox = new JComboBox(new String[] {"Fruchterman-Reingold-Layout", "KamadaKawaiLayout",});
			layoutBox.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					if (layoutBox.getSelectedItem().equals("Fruchterman-Reingold-Layout")) {
						((Canvas) canvas).doInit();
						((Canvas) canvas).forceDirectedLayout();
						((Canvas) canvas).repaint();
					} else {
//						((Canvas) canvas).doInit();
//						((Canvas) canvas).KKLayout();//TODO
						Container c = getCanvas(dimension).getParent();
						while( !(c instanceof DisChoco) ){
							c = c.getParent();
						}
						((DisChoco) c).getConsole().printlnWarning("Not yet Implemented");
//						((Canvas) canvas).repaint();
					}
				}
			});
		}
		layoutBox.setSelectedItem("Fruchterman-Reingold-Layout");
		return layoutBox;
	}

	/**
	 * @param d
	 * @return
	 */
	public JPanel getCanvas(Dimension d) {
		if (canvas == null) {
			dimension = d;
			canvas = new Canvas(d);
		}
		return canvas;
	}

	/**
	 * This method initializes browse
	 * 
	 * @return javax.swing.JButton
	 */
	public JButton getZoomIn() {
		if (zoomIn == null) {
			zoomIn = new JButton();
			zoomIn.setText(" + ");
			zoomIn.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					((Canvas) canvas).zoom(1.1f);
					Container c = getParent();
					while (!(c instanceof JFrame)){
						c = c.getParent();
					}
					((DisChoco) c).getConsole().printlnWarning("Zoom + TODO");
//					((Canvas) canvas).scale(1.1f);
				}
			});
		}
		return zoomIn;
	}

	/**
	 * This method initializes browse
	 * 
	 * @return javax.swing.JButton
	 */
	public JButton getZoomOut() {
		if (zoomOut == null) {
			zoomOut = new JButton();
			zoomOut.setText(" - ");
			zoomOut.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					((Canvas) canvas).zoom(1/1.1f);
					Container c = getParent();
					while (!(c instanceof JFrame)){
						c = c.getParent();
					}
					((DisChoco) c).getConsole().printlnWarning("Zoom - TODO");
//					((Canvas) canvas).scale(1/1.1f);
				}
			});
		}
		return zoomOut;
	}

	@SuppressWarnings("serial")
	public class Canvas extends JPanel implements MouseListener, MouseMotionListener, MouseWheelListener {

		public static final double DEFAULT_ZOOM_PERCENTAGE = 1.0;
		private double minX, maxX, minY, maxY;
		private Point2D center;
		private double scale = 1.0;

		/**
		 * This method initializes
		 * 
		 * @param height
		 * @param width
		 * @return javax.swing.JPanel
		 */
		public Canvas(Dimension d) {
			minimumSize = new Dimension(d.width-25, d.height - 100);
			maximumSize = new Dimension((minimumSize.width * MAX_ZOOM), minimumSize.height * MAX_ZOOM);
			preferredSize = minimumSize;
			max_dimension = Math.max(preferredSize.width, preferredSize.height);
			setBackground(Color.white);
			setSize(preferredSize);
			doInit();
			forceDirectedLayout();
			setPreferredSize(preferredSize);
			addMouseWheelListener(this);
			addMouseListener(this);
		}

		public void paint(Graphics g) {
			Graphics2D gr = (Graphics2D) g;
			int w = getWidth();
			int h = getHeight();
			super.paint(g);
//			calculateCenter();
//			double cx = center.getX();
//			double cy = center.getY();
//			AffineTransform tx = new AffineTransform();
//			tx.translate(cx/2, cy/2);
//			tx.scale(scale, scale);
//			tx.translate(-cx/2, -cy/2);
//			gr.setTransform(tx);
//			centering();
			gr.setRenderingHints(hints);		
			gr.setColor(Color.white);
			m = gr.getFontMetrics();
			gr.fillRect(0, 0, w, h);
			for (Edge edge : graph.getEdges()) {
				edge.draw(g);
			}
			for (Vertex v : graph.getVertices()) {
				v.draw(g, m);
			}
		}
		
		public void scale(double s){
			double ss = s*scale;
			if(ss < 0.5 || ss > MAX_ZOOM){
				return;
			}
			scale = ss;
			calculateCenter();
			double cx = center.getX();
			double cy = center.getY();
			for (Vertex v : graph.getVertices()) {
				double dx = (cx - v.x);
				double dy = (cy - v.y);
				v.x += dx - s*(dx);
				v.y += dy - s*(dy);
//				v.scale(scale);
			}
			repaint();
		}

		public void mouseClicked(MouseEvent e) {}

		public void mouseEntered(MouseEvent e) {}

		public void mouseExited(MouseEvent e) {}

		public void mousePressed(MouseEvent e) {
			Point p = e.getPoint();
			addMouseMotionListener(this);
			if (PICKING) {
				for (Vertex v : graph.getVertices()) {
					if (v.contains(p)) {
						pick = v;
						break;
					}
				}
				if (pick != null) {
					setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
					pickfixed = pick.selected;
					pick.selected = true;
					pick.setCenter(p);
				}
			} else { 
				setCursor(Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR));
				double x = p.x - center.getX(); // previewX
				double y = p.y - center.getY();
				for (Vertex v : graph.getVertices()) {
					v.x += x;
					v.y += y;
				}
				center.setLocation(p);
			}
			repaint();
			e.consume();
		}

		public void mouseDragged(MouseEvent e) {
			Point p = e.getPoint();
			if (PICKING) {
				if (pick != null) {
					setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
					pick.setCenter(p);
				}
			} else {
				setCursor(Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR));
				double x = p.x - center.getX(); // previewX
				double y = p.y - center.getY();
				for (Vertex v : graph.getVertices()) {
					v.x += x;
					v.y += y;
				}
				center.setLocation(p);
			}
			repaint();
			e.consume();
		}

		public void mouseReleased(MouseEvent e) {
			Point p = e.getPoint();
			if (PICKING) {
				removeMouseMotionListener(this);
				if (pick != null) {
					pick.setCenter(p);
					pick.selected = pickfixed;
					pick = null;
					repaint();
					e.consume();
				}
			} else {
				center.setLocation(p);
			}
			setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
		}

		public void mouseMoved(MouseEvent e) {}

		public void mouseWheelMoved(MouseWheelEvent e) {
			updatePreferredSize(e.getWheelRotation(), e.getPoint());
		}

		public Dimension getPreferredSize() {
			return preferredSize;
		}

		private void updatePreferredSize(int n, Point p) {
			double d = (double) n;
			d = (n > 0)? (1 / d) : (-d);
			zoom(d);
		}

		private void zoom(double d) {
			int w = (int) ((getWidth()) * d * DEFAULT_ZOOM_PERCENTAGE);
			int h = (int) ((getHeight()) * d * DEFAULT_ZOOM_PERCENTAGE);
			if (w < minimumSize.getWidth()) {
				w = minimumSize.width;
			} else if (w > maximumSize.getWidth()) {
				w = maximumSize.width;
			}
			if (h < minimumSize.getHeight()) {
				h = minimumSize.height;
			} else if (h > maximumSize.getHeight()) {
				h = maximumSize.height;
			}
			preferredSize = new Dimension(w, h);
			max_dimension = Math.max(w, h);
			doLayout();
			revalidate();
			repaint();
		}

		public void forceDirectedLayout() {
			while (!done()) {
				step();
			}
			calculateCenter();
			centering();
		}

		/**
		 * 
		 */
		private void centering() {
			int w = getWidth()/2;
			int h = getHeight()/2;
			for(Vertex v : graph.getVertices()){
				v.x += (w-center.getX());
				v.y += (h-center.getY());
			}
		}

		/**
		 * 
		 */
		private void calculateCenter() {
			minX = getWidth();
			minY = getHeight();
			maxX = 0;
			maxY = 0;
			for(Vertex v : graph.getVertices()){
				double x = v.getX();
				if (x < minX) {
					minX = x;
				} else if (x > maxX) {
					maxX = x;
				}
				double y = v.getY();
				if (y < minY) {
					minY = y;
				} else if (y > maxY) {
					maxY = y;
				}
			}
			center = new Point2D.Double(minX+(maxX-minX)/2, minY+(maxY-minY)/2);
		}

		private void doInit() {
			if (graph != null) {
				currentIteration = 0;
				temperature = getWidth() / 10;
				forceConstant = Math.sqrt(getHeight() * getWidth() / graph.getVertexCount());
				attraction_constant = attraction_multiplier * forceConstant;
				repulsion_constant = repulsion_multiplier * forceConstant;
			}
		}

		/**
		 * Moves the iteration forward one notch, calculation attraction and repulsion between vertices and edges and cooling the temperature.
		 */
		public synchronized void step() {
			currentIteration++;
			/** Calculate repulsion */
			calculateRepulsions();
			/** Calculate attraction */
			calculateAttractions();
			calculatePositions();
			cool();
		}

		/**
		 * calculate repulsive forces
		 */
		private void calculateRepulsions() {
			for (Vertex v : graph.getVertices()) {
				v.dx = 0;
				v.dy = 0;
				for (Vertex u : graph.getVertices()) {
					if (v.id != u.id) {
						double delta_x = v.getX() - u.getX();
						double delta_y = v.getY() - u.getY();
						double dist = Math.sqrt(delta_x * delta_x + delta_y * delta_y);
						double deltaLength = Math.max(EPSILON, dist);
						double force = (repulsion_constant * repulsion_constant) / deltaLength;
						v.dx += (delta_x / deltaLength) * force;
						v.dy += (delta_y / deltaLength) * force;
					}
				}
			}
		}

		/**
		 * calculate attractive forces
		 */
		private void calculateAttractions() {
			for (Edge e : graph.getEdges()) {
				Vertex v = e.source;
				Vertex u = e.target;
				double delta_x = v.getX() - u.getX();
				double delta_y = v.getY() - u.getY();
				double dist = Math.sqrt(delta_x * delta_x + delta_y * delta_y);
				double deltaLength = Math.max(EPSILON, dist);
				double force = (deltaLength * deltaLength) / attraction_constant;
				if (!Double.isNaN(force)) {
					v.dx -= (delta_x / deltaLength) * force;
					v.dy -= (delta_y / deltaLength) * force;
					u.dx += (delta_x / deltaLength) * force;
					u.dy += (delta_y / deltaLength) * force;
				}
			}
		}

		private void calculatePositions() {
			for (Vertex v : graph.getVertices()) {
				double dist = Math.sqrt(v.dx * v.dx + v.dy * v.dy);
				double deltaLength = Math.max(EPSILON, dist);
				double disp_x = v.dx / deltaLength * Math.min(deltaLength, temperature);
				double disp_y = v.dy / deltaLength * Math.min(deltaLength, temperature);
				v.x += disp_x;
				v.y += disp_y;

				double borderWidth = getWidth() / BORDER;
				double x = v.x;
				if (x < borderWidth) {
					x = borderWidth + Math.random() * borderWidth * 2.0;
				} else if (x > (getWidth() - borderWidth)) {
					x = getWidth() - borderWidth - Math.random() * borderWidth * 2.0;
				}

				double borderHeight = getHeight() / BORDER;
				double y = v.y;
				if (y < borderHeight) {
					y = borderHeight + Math.random() * borderHeight * 2.0;
				} else if (y > (getHeight() - borderWidth)) {
					y = getHeight() - borderHeight - Math.random() * borderHeight * 2.0;
				}
				v.x = x;
				v.y = y;
			}
		}

		/**
		 * reduce the temperature as the layout approaches a better configuration
		 */
		private void cool() {
			temperature *= (1.0 - currentIteration / (double) maxIterations);
		}

		/**
		 * Returns true once the current iteration has passed the maximum count, <tt>MAX_ITERATIONS</tt>.
		 */
		public boolean done() {
			return (currentIteration > maxIterations || temperature < 1.0 / max_dimension);
		}
	}

	/**
	 * Creates a triangle shape that points downwards.
	 * 
	 * @param s
	 *            the size factor (equal to half the height of the triangle).
	 * @return A triangle shape.
	 */
	public static Shape createDownTriangle(final float s) {
		final GeneralPath p0 = new GeneralPath();
		p0.moveTo(0.0f, s);
		p0.lineTo(s, -s);
		p0.lineTo(-s, -s);
		p0.closePath();
		return p0;
	}

	/**
	 * Creates a triangle shape that points upwards.
	 * 
	 * @param s
	 *            the size factor (equal to half the height of the triangle).
	 * @return A triangle shape.
	 */
	public static Shape createUpTriangle(final float s) {
		final GeneralPath p0 = new GeneralPath();
		p0.moveTo(0.0f, -s);
		p0.lineTo(s, s);
		p0.lineTo(-s, s);
		p0.closePath();
		return p0;
	}

	/**
	 * Creates a diamond shape.
	 * 
	 * @param s
	 *            the size factor (equal to half the height of the diamond).
	 * @return A diamond shape.
	 */
	public static Shape createDiamond(final float s) {
		final GeneralPath p0 = new GeneralPath();
		p0.moveTo(0.0f, -s);
		p0.lineTo(s, 0.0f);
		p0.lineTo(0.0f, s);
		p0.lineTo(-s, 0.0f);
		p0.closePath();
		return p0;
	}

	public static Shape createSensor(int x, int y) {
		int width = 32, height = 32;
		GeneralPath shape = new GeneralPath();
		shape.moveTo(x - width / 2, y - height / 2);
		shape.moveTo(10, 10);
		shape.lineTo(20, 20);
		shape.quadTo(30, 50, 160, 30);
		shape.curveTo(50, 150, 260, 60, 70, 20);
		shape.closePath();
		return shape;
	}

	@SuppressWarnings("serial")
	public static void main(String[] args) {
		javax.swing.JFrame frame = new javax.swing.JFrame();
		frame.setVisible(true);
		frame.setSize(400, 400);
		JPanel container = new JPanel() {
			public void paint(Graphics g) {
				GeneralPath shape = new GeneralPath();
				shape.moveTo(10, 10);
				shape.lineTo(20, 20);
				shape.quadTo(30, 50, 160, 30);
				shape.curveTo(50, 150, 260, 60, 70, 20);
				shape.closePath();
				Graphics2D gr = ((Graphics2D) g);
				gr.setBackground(Color.white);
				gr.setColor(Color.black);
				gr.setStroke(new BasicStroke(2));
				gr.draw(shape);
			}
		};
		frame.setContentPane(container);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	}
}
