package kernel.constraints.integer;

import java.util.Iterator;
import java.util.List;

import kernel.vars.integer.IntDomainIterator;
import kernel.vars.integer.IntExternalVar;
import kernel.vars.integer.IntVariable;

public class ExtensionBinConstraint extends SIntBinConstraint {

	public BinRelation relation;

	public ExtensionBinConstraint(String type, IntVariable var0, IntExternalVar var1, List<int[]> mat) {
		super(var0, var1);
		relation = makeBinaryRelation(var0, var1, mat, type.equals(feasP));
		this.type = type;
	}

	public static BinRelation makeBinaryRelation(IntVariable v0, IntVariable v1, List<int[]> mat, boolean feas) {
		int n0 = v0.getSup() - v0.getInf() + 1;
		int n1 = v1.getSup() - v1.getInf() + 1;
		int inf0 = v0.getInf();
		int inf1 = v1.getInf();
		BinRelation relation;
		relation = new CouplesTable(feas, inf0, inf1, n0, n1);
		Iterator<int[]> it = mat.iterator();
		while (it.hasNext()) {
			int[] tuple = (int[]) it.next();
			if (tuple.length != 2)
				throw new Error("Wrong dimension : " + tuple.length + " for a tuple");
			((CouplesTable) relation).setCouple(tuple[0], tuple[1]);
		}
		return relation;
	}

	/**
	 * Returns a string representation of this object
	 */
	public String toString() {
		return "C_" + this.getIndex() + "[ " + this.v0.toString() + "@" + type + "@" + this.v1.toString() + "]  ";
	}

	public boolean isConsistent(int val0, int val1) {
		this.incrementNumConstraintChecks(1);
		return this.relation.isConsistent(val0, val1);
	}

	@Override
	public void propagate() {
		awakeOnVar(v0);
		awakeOnVar(v1);
	}

	public void awakeOnVar(IntVariable v) {
		IntDomainIterator it = v.getDomain().iterator();
		while (it.hasNext()) {
			int value = it.next();
			int support = this.findSupport(value, v);
			if (support == Integer.MIN_VALUE) {
				v.removeValue(value, index);
			}
		}
	}

	/**
	 * @param value
	 */
	public int getNumberOfSupport(int value) {
		return this.relation.getNumberOfSupports(value);
	}

	/**
	 * @param value
	 * @return
	 */
	public int getNumberOfConflicts(int value) {
		return this.relation.getNumberOfConflicts(value);
	}

}
