package kernel.events;

import java.util.SortedSet;
import java.util.TreeSet;

import kernel.Constraint;
import kernel.vars.integer.IntVar;
import util.BitSet;
import util.IntArray;

public class IntVariableEvent extends VariableEvent {

	public IntVariableEvent(IntVar var) {
		super(var);
	}

	@Override
	public void propagateOneBasicEvent() {
		if (this.eventType != EMPTYEVENT) {
			int eventT = this.eventType;
			int evCause = this.cause;

			this.eventType = EMPTYEVENT;
			this.cause = NOEVENT;

			int mostImportantEvent = BitSet.lowestBit(eventT);

			eventT = BitSet.unsetBit(eventT, mostImportantEvent);

			switch (mostImportantEvent) {
				case DESINSTANCIATE :
					propagateEvent(DESINSTANCIATE, NOCAUSE);
				break;
				case INSTANCIATE :
					propagateEvent(INSTANCIATE, cause);
				break;
				case INCREMENT_INF :
					propagateRemEvent(INCREMENT_INF, cause);
				break;
				case DECREMENT_SUP :
					propagateRemEvent(DECREMENT_SUP, cause);
				break;
				case REMOVE_VAL :
					propagateRemEvent(REMOVE_VAL, cause);
				break;
				default :
					System.err.println("calls of propagate one event on empty event! ");
			}
			this.eventType |= eventT;
			if (this.cause == NOEVENT && this.eventType != EMPTYEVENT) {
				this.cause = evCause;
			}
		}
	}

	protected void propagateRemEvent(int evtType, int evtCause) {
		((IntVar) this.touchedVar).getDomain().freezeDeltaDomain();
		this.propagateEvent(evtType, evtCause);
		((IntVar) this.touchedVar).getDomain().releaseDeltaDomain();
	}

	protected void propagateEvent(int evtType, int evtCause) {
		this.constraintHandler.notifyPropagateVarEvent(this.touchedVar, evtType);
		IntArray indices = this.touchedVar.getListeners(evtType);

		SortedSet<Constraint> listeners = new TreeSet<Constraint>();
		// Firstly we sort the constraints
		for (int index : indices) {
			Constraint c = this.constraintHandler.getConstraint(index);
			if (c == null)
				System.err.println("" + index + " ###################### " + this.constraintHandler);
			listeners.add(c);
		}
		// //Secondly get the sorted indices
		// indices = new IntArray();
		// Iterator<Constraint> it = constraints.iterator();
		// while (it.hasNext()) {
		// indices.add(it.next().getIndex());
		// }

		// Propagate following the priority ordering
		for (Constraint c : listeners) {
			if (c != null && c.getIndex() != evtCause) {
				// if (c.getIndex() != evtCause) {
				c.awakeOnVariableEvent(touchedVar, evtType);
			}
		}
	}

	// ====================================================================
	/**
	 * Clears the variable event : delegates to the variable event.
	 */
	public void clear() {
		this.eventType = EMPTYEVENT;
		this.cause = NOEVENT;
		((IntVar) this.touchedVar).getDomain().clearDeltaDomain();
	}
}
