/************************************************************
ABT add permanent link
 *************************************************************/

package simple.dcsp.abtfamily.abt;

import java.util.Iterator;

import kernel.Communicator;
import kernel.StopException;
import kernel.com.Message;
import kernel.events.AgentEventListener;
import kernel.events.VariableEvent;
import kernel.vars.integer.IntExternalVar;
import simple.Nogood;
import simple.SingleExplainedVarAgent;
import simple.dcsp.abtfamily.ABT;
import simple.dcsp.abtfamily.IdleEvent;
import simulation.metric.Metric;
import util.Commands;
import util.IntArray;
import util.VarAssignment;
import util.VarEventMask;

public class ABTAgent extends SingleExplainedVarAgent implements ABT {

	public OkMessageHandler okMsgH = null;
	public NogoodsHandler ngdMsgH = null;
	public AddLinkMsgHandler adlMsgH = null;

	/**
	 * @param agentId
	 * @param agName
	 * @param description
	 */
	public ABTAgent(int agentId, String agName, String description) {
		super(agentId, agName, description);
		this.constructCommunicator();
		this.terminateMsgHandler = new ABTTermDetection(this);
		// -------------- Messages Handler ------------------//
		this.okMsgH = new OkMessageHandler(this);
		this.ngdMsgH = new NogoodsHandler(this);
		this.adlMsgH = new AddLinkMsgHandler(this);
		this.initAuxStatistics();
	}

	private void initAuxStatistics() {
		this.auxiliaryStatistics.put(Commands.nbObsoleteNogood, new Metric(Commands.nbObsoleteNogood));
		this.auxiliaryStatistics.put(Commands.nbAssignments, new Metric(Commands.nbAssignments));
		this.auxiliaryStatistics.put(Commands.ObsoleteOk, new Metric(Commands.ObsoleteOk));
		this.auxiliaryStatistics.put(Commands.MSGsSIZE, new Metric(Commands.MSGsSIZE));
		this.auxiliaryStatistics.put(Commands.MaxMSGsSIZE, new Metric(Commands.MaxMSGsSIZE, 0L, Metric.max));
	}

	/**
	 * @param msg
	 */
	public void storeStatisticsFor(Message msg) {
		long size = msg.getSize();
		this.auxiliaryStatistics.get(Commands.MSGsSIZE).value += size;
		if(this.auxiliaryStatistics.get(Commands.MaxMSGsSIZE).value < size){
			this.auxiliaryStatistics.get(Commands.MaxMSGsSIZE).value = size;
		}
	}

	/**
	 * 
	 */
	private void constructCommunicator() {
		this.communicator = new Communicator(this) {
			protected boolean waitForArrivalMessages(long msTimeout) throws InterruptedException {
				logDomains();
				this.waitingForMessage = true;
				boolean arrived = false;
				if (((ABT) this.myAgent).isIdle()) {
					arrived = mailBoxLock.waitUntilTrue(10);
					if (!arrived)
						this.myAgent.fireAgentEvent(new IdleEvent(this.myAgent));
				}
				if (!arrived) {
					arrived = mailBoxLock.waitUntilTrue(msTimeout);
				}
				this.waitingForMessage = false;
				logDomains();
				return arrived;
			}
		};
	}

	// ==========================================================================
	@Override
	public void initAgentView() {
		IntArray agents = this.communicator.getHigherNeighbors();
		for (int agId : agents) {
			this.agentView.add((IntExternalVar) this.getExternalVar(agId));
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see AbstractAgent#addMsgHandlers()
	 */
	@Override
	public void addMsgHandlers() {
		this.communicator.addMsgHandler(this.okMsgH);
		this.communicator.addMsgHandler(this.ngdMsgH);
		this.communicator.addMsgHandler(this.adlMsgH);
	}

	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.DESINSTANCIATE);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_THE_LAST).add(VariableEvent.RESTORE_VAL);
		this.eventsConfig.get(VarEventMask.LOCAL_VAR_IS_BETWEEN_L_H).add(VariableEvent.RESTORE_VAL);
		this.constraintHandler.update();
	}

	// ////////////////////////////////////////////////////////////////////
	// //////////////////////////// ABT Procedures: //////////////////////
	// ///////////////////////////////////////////////////////////////////
	@Override
	public void startDistributedSearch() throws StopException {
		this.addAgentEventListener((AgentEventListener) this.terminateMsgHandler);
		((ABTTermDetection) this.terminateMsgHandler).sendInitialEndAgentMessage();

		this.checkAgentView();
	}

	/**
	 * choose an instantiation (i.e a value) compatible with the agentView
	 * 
	 * @throws StopException
	 */
	public void checkAgentView() throws StopException {
		int myValue = this.var.getValue();
		if (!this.var.getDomain().contains(myValue)) {
			if (this.valSelector.hasNextValue()) {
				int value = this.valSelector.getNextValue();
				this.var.instantiate(value);
				this.okMsgH.sendOkMessages();
			} else {
				this.backtrack();
			}
		}
	}

	public boolean isIdle() {
		if (this.disSearchStarted) {
			IntArray higherNeighbors = this.communicator.getHigherNeighbors();
			return this.var.isInstantiated() && this.agentView.toInstantiation().getAgentIds().contains(higherNeighbors);
		}
		return false;
	}

	// /////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/**
	 * @throws StopException
	 */

	public void backtrack() throws StopException {
		this.resolveContradiction();
		this.checkAgentView();
	}

	// /////////////////////////////////////////////////////////////////////////////////////////////////////////////

	public void resolveContradiction() throws StopException {
		Nogood nogood = this.solveMyNogoods();
		if (nogood.isEmpty()) { throw new StopException(Commands.FAILURE); }
		this.agentView.clear(nogood.getRhs());
		this.ngdMsgH.sendNogood(nogood);
	}

	// /////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/**
	 * Procedure to check if a new link should be added.
	 * 
	 * @param nogood
	 */
	public void checkAddLink(Nogood nogood) {
		Iterator<VarAssignment> it = nogood.getLhs().iterator();
		while (it.hasNext()) {
			VarAssignment ass = it.next();
			if (!this.communicator.isHigherNeighbor(ass.getAgentId())) {
				this.communicator.addNeighbor(ass.getAgentId());
				this.agentView.add(this.makeExternalVar(ass));
				this.adlMsgH.sendAddLink(ass);
			}
		}
	}

	@Override
	public void sendOkMessage(int agId) {
		this.okMsgH.sendOkMessage(agId);
	}

}
