package simple.dcsp.dba;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;

import kernel.AbstractAgent;
import kernel.AbstractMaster;
import kernel.com.Message;
import kernel.order.AgentComplexity;
import kernel.order.LexicographicAO;
import util.Commands;
import util.graph.Graph;

public class DBAmaster extends AbstractMaster {

	public DBAmaster() {
		this.setCentralizedAO(new LexicographicAO());
	}

	/**   
	 */
	public void broadcast(Message msg, int agentId) {
		if (msg.getType().equals(Commands.AO)) {
			Graph<Integer> graph = new Graph<Integer>();
			HashMap<Integer, AgentComplexity> agCmpx = this.centralizedAO.getAgentsComplexity();
			for (int v : agCmpx.keySet()) {
				graph.addVertex(new Integer(v));
				AgentComplexity neig = agCmpx.get(v);
				for (int w : neig.getNeighbors()) {
					graph.addEdge(v, w);
				}
			}
			for (int id : this.centralizedAO.getOrder()) {
				Message message = msg.clone();
				message.addContents("Cyclic?", graph.isCyclic() + "");
				message.addContents("Max-Distance", graph.maxDistance(id) + "");
				message.setReceiverId(id);
				this.sendMessage(message);
			}
		}
	}

	@Override
	public AbstractAgent makeAgent(int agentId, String agentName, String description) {
		return new DBAagent(agentId, agentName, description);
	}

	/* (non-Javadoc)
	 * @see kernel.AbstractMaster#getTypesOfMessages()
	 */
	@Override
	public Collection<String> getTypesOfMessages() {
		Collection<String> ans = new ArrayList<String>();
		ans.add(Commands.OK);
		ans.add(Commands.IMPROVE);
		ans.add(Commands.TERMINATION_DETETION);
		return ans;
	}


}
