package simple.dcsp.sbt;

import kernel.StopException;
import kernel.VarIdentifier;
import kernel.events.VariableEvent;
import kernel.vars.integer.IntExternalVar;
import simple.SimpleAgent;
import util.Commands;
import util.Instantiation;
import util.IntArray;
import util.VarAssignment;
import util.VarEventMask;

/**
 * Implementation of the improved version of SBT;
 * 
 * @author Wahbi Mohamed
 * 
 */
public class SBTAgent extends SimpleAgent {

	protected CPAMsgeHandler cpaMsgH = null;
	protected BackCPAMsgHandler backCPAMsgH = null;

	/**
	 * Constructors
	 * 
	 * @param description
	 * @param agentName
	 */
	public SBTAgent(int agentId, String agName, String description) {
		super(agentId, agName, description);
		this.constraintHandler = new SBTConstraintsHandler(this);
		this.communicator = new SBTCommunicator(this);
		// -------------- Messages Handler ------------------//
		this.terminateMsgHandler = new SBTTerminateMsgHandler(this);
		this.cpaMsgH = new CPAMsgeHandler(this);
		this.backCPAMsgH = new BackCPAMsgHandler(this);
	}

	// /////////////////////////////////////////////////////////////////////////////////////////////////

	@Override
	public void configurePropagationEngine() {
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.INSTANCIATE);
		this.eventsConfig.get(VarEventMask.EXTERNAL_VAR_HIGHER_THAN_LOCAL_VAR).add(VariableEvent.DESINSTANCIATE);
		this.constraintHandler.update();
	}

	// ==========================================================================

	public void initAgentView() {
		IntArray agents = this.agentOrder.getHigherAgents();
		for (int agId : agents) {
			if (!this.communicator.isNeighbor(agId)) {
				VarAssignment var = new VarAssignment(new VarIdentifier(agId), Integer.MIN_VALUE);
				IntExternalVar extVar = this.makeExternalVar(var);
				this.agentView.add(extVar);
			} else
				this.agentView.add((IntExternalVar) this.getExternalVar(agId));
		}
	}

	// ==========================================================================
	/**
	 * 
	 */
	public void addMsgHandlers() {
		this.communicator.addMsgHandler(this.cpaMsgH);
		this.communicator.addMsgHandler(this.backCPAMsgH);
	}

	// /////////////////////////////////////////////////////////////////////////////////////////////////
	/**
	 * Start distributed search from the initializing agent
	 * 
	 * @throws StopExeption
	 */
	public void startDistributedSearch() throws StopException {
		// if(emptyDomain)
		// throw new StopException(Commands.FAILURE);
		if (this.agentOrder.getRankOf(this.agentId) == 0) {
			this.assignCPA();
		}
	}

	// /////////////////////////////////////////////////////////////////////////////////////////////////
	/**
	 * this function is called when ever we need to instantiate the current partial assignment, if doesn't succeed it launches a backtrack event
	 */
	public void assignCPA() throws StopException {
		if (this.valSelector.hasNextValue()) {
			int val = this.valSelector.getNextValue();
			this.var.instantiate(val);
			Instantiation cpa = this.agentView.toInstantiation();
			int nextAgentId = this.getNextAgent();
			if (nextAgentId == -1) {
				throw new StopException(Commands.SUCCESS);
			} else {
				cpa.add(this.var.getVarAssignment());
				this.cpaMsgH.sendCPAMessage(cpa, nextAgentId);
			}
		} else {
			this.backtrack();
		}
	}

	private int getNextAgent() {
		return this.agentOrder.getNextOf(this.agentId);
	}

	// /////////////////////////////////////////////////////////////////////////////////////////////////
	/**
	 * called when contradiction is generated.
	 * 
	 * @throws StopException
	 */
	public void backtrack() throws StopException {
		int agId = this.agentOrder.getPreviousOf(this.agentId);
		if (agId == -1) {
			throw new StopException(Commands.FAILURE);
		} else {
			this.backCPAMsgH.sendBack(agId);
			this.agentView.clear();
			this.getPropagationEngine().flushEvents();
			this.var.getDomain().reset();

		}
	}

	public void processCPA(Instantiation cpa) throws StopException {
		this.agentView.priorityMerge(cpa);
		if (var.getDomainSize() == 0) {
			this.backtrack();
		} // FC
		else
			this.assignCPA();
	}

	public void proceeeBack() throws StopException {
		int myVal = var.getValue();
		var.removeValue(myVal, VariableEvent.NOCAUSE);
		assignCPA();
	}

}
